@@ -26,22 +26,24 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
-from enum import Enum
-import logging
-import hashlib
 import collections.abc as abc
-from dataclasses import is_dataclass, fields
-
-# Removing this in 2020-12 broke a shocking amount of stuff, such as
-# https://github.com/OP2/PyOP2/pull/605
-# Bring it back for now to mitigate the breakage, however this is going
-# away in 2021 at the latest.
-new_hash = hashlib.sha256
-
+import errno
+import hashlib
+import logging
 import os
 import shutil
 import sys
-import errno
+from dataclasses import fields as dc_fields, is_dataclass
+from enum import Enum
+
+
+try:
+    import attrs
+except ModuleNotFoundError:
+    _HAS_ATTRS = False
+else:
+    _HAS_ATTRS = True
+
 
 logger = logging.getLogger(__name__)
 
@@ -236,9 +238,14 @@ class KeyBuilder:
             try:
                 method = getattr(self, "update_for_"+tname)
             except AttributeError:
-                # Handling numpy >= 1.20, for which
-                # type(np.dtype("float32")) -> "dtype[float32]"
-                if tname.startswith("dtype[") and "numpy" in sys.modules:
+                if (
+                        # Handling numpy >= 1.20, for which
+                        # type(np.dtype("float32")) -> "dtype[float32]"
+                        tname.startswith("dtype[")
+                        # Handling numpy >= 1.25, for which
+                        # type(np.dtype("float32")) -> "Float32DType"
+                        or tname.endswith("DType")
+                        ) and "numpy" in sys.modules:
                     import numpy as np
                     if isinstance(key, np.dtype):
                         method = self.update_for_specific_dtype
@@ -249,6 +256,9 @@ class KeyBuilder:
                 elif is_dataclass(tp):
                     method = self.update_for_dataclass
 
+                elif _HAS_ATTRS and attrs.has(tp):
+                    method = self.update_for_attrs
+
             if method is not None:
                 inner_key_hash = self.new_hash()
                 method(inner_key_hash, key)
@@ -343,7 +353,14 @@ class KeyBuilder:
     def update_for_dataclass(self, key_hash, key):
         self.rec(key_hash, type(key_hash).__name__.encode("utf-8"))
 
-        for fld in fields(key):
+        for fld in dc_fields(key):
+            self.rec(key_hash, fld.name)
+            self.rec(key_hash, getattr(key, fld.name, None))
+
+    def update_for_attrs(self, key_hash, key):
+        self.rec(key_hash, type(key_hash).__name__.encode("utf-8"))
+
+        for fld in attrs.fields(key.__class__):
             self.rec(key_hash, fld.name)
             self.rec(key_hash, getattr(key, fld.name, None))
 
@@ -533,12 +550,13 @@ class _PersistentDictBase:
 
     @staticmethod
     def _write(path, value):
-        from pickle import dump, HIGHEST_PROTOCOL
+        from pickle import HIGHEST_PROTOCOL, dump
         with open(path, "wb") as outf:
             dump(value, outf, protocol=HIGHEST_PROTOCOL)
 
     def _item_dir(self, hexdigest_key):
         from os.path import join
+
         # Some file systems limit the number of directories in a directory.
         # For ext4, that limit appears to be 64K for example.
         # This doesn't solve that problem, but it makes it much less likely
